#!/usr/bin/env python
#file ngs_variant_mapper.py

# Copyright (C) 2008 James Smagala

# Contact:
#   Sujatha Seenu
#   sujvel@gmail.com

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

""" Map nucleiotide sequence variations in the consensus sequence """


import wx
import os
import re
import sys
import time
import copy
import datetime
from itertools import groupby


from gui.ngs_variant_mapper_gui import VarMapGUIMainFrame
from core.easy_addcol_xls import read_excel
from core.sequence import format_fasta


########################################################################

wildcard = "All files (*.*)|*.*"

########################################################################




class VarMapError(Exception):
    """Custom error class for specific errors"""
    pass


def pick_dir(parent=None, current_dir=''):
    """Select a directory"""
    dlg = wx.DirDialog(parent, "Select a directory:", defaultPath=current_dir)
    dir = current_dir
    if dlg.ShowModal() == wx.ID_OK:
        # this dialog was not canceled
        dir = dlg.GetPath()
    dlg.Destroy()
    return dir


def error_dialog(parent=None, message='An unknown error occured'):
    """Display an error dialog"""
    message = unicode(message)
    dlg = wx.MessageDialog(parent, message, caption='Error',
                           style=wx.OK|wx.ICON_ERROR)
    dlg.ShowModal()
    #dlg.Destroy()


class VarMapMainFrame(VarMapGUIMainFrame):
    """Inherit from wxGlade-generated frame and add functionality"""
    def __init__(self, *args, **kwargs):
        super(VarMapMainFrame, self).__init__(*args, **kwargs)

        # setup an icon for the window
        icon = wx.Icon('icons/ngs_var_map.ico', wx.BITMAP_TYPE_ICO)
        self.SetIcon(icon)

        # get a reference to the app object
        self.app = wx.GetApp()

        # bind events to handlers
        self.Bind(wx.EVT_CLOSE, self.OnExit)

    def OnGetDir1(self, event):
        """Get the consensus sequence file"""
        dlg = wx.FileDialog(
            self, message="Choose a file",
            defaultFile="",
            wildcard=wildcard,
            style=wx.OPEN | wx.MULTIPLE | wx.CHANGE_DIR
            )
        if dlg.ShowModal() == wx.ID_OK:
            self.dir1.SetValue(''.join(dlg.GetPaths()))

        dlg.Destroy()

    def OnGetDir2(self, event):
        """Get the variant call file"""
        dlg = wx.FileDialog(
            self, message="Choose a file",
            defaultFile="",
            wildcard=wildcard,
            style=wx.OPEN | wx.MULTIPLE | wx.CHANGE_DIR
            )
        if dlg.ShowModal() == wx.ID_OK:
            self.dir2.SetValue(''.join(dlg.GetPaths()))

        dlg.Destroy()

    def OnRunVarMap(self, event):
        """Get variant sequences"""
        con_file = self.dir1.GetValue()
        map_file = self.dir2.GetValue()
        #Check for a valid input
        if 'xls' not in map_file:
            wx.MessageBox(' ** Input File Error ** Check your input files. ','Info',wx.OK | wx.ICON_INFORMATION)
        #Status bar
        self.dialog=wx.ProgressDialog('Job Status', 'Process is running', maximum=100, parent=None,style=wx.PD_AUTO_HIDE|wx.PD_APP_MODAL)



        # Output file
        seqid=position=base=link=frequency=insertion=""
        map_seq={}
        linkage={}
        con_fasta={}
        variant={}
        ln_pos=[]
        ln_list=[]
        pb2_seq=pb1_seq=pa_seq=ha_seq=np_seq=na_seq=m_seq=[]
        #Parse the mapped variants spreadsheet
        data,max_row,max_col=read_excel(map_file)[0][1:]
        for row in range(max_row):
            if data[(row,2)]:
                try:
                    seqid=data[(row,0)].strip()
                    cdcid=data[(row,0)].split('_')[0].strip()
                except:
                    pass
                try:
                    position=str(data[(row,1)]).strip()
                except:
                    pass
                try:
                    base=data[(row,5)].strip()
                except:
                    pass
                try:
                    ref=data[(row,4)].strip()
                except:
                    pass
                try:
                    link=str(data[(row,6)]).strip()
                except:
                    pass
                try:
                    frequency=str(data[(row,10)]).strip()
                except:
                    pass
                try:
                    insertion=''
                    if 'Insertion' in str(data[(row,2)]):
                        insertion=str(data[(row,2)])
                except:
                    pass
                seq_key=seqid+"="+position.strip()+"="+link+"="+frequency+"="+insertion
                map_seq[seq_key]=base+"="+str(len(ref))
                if link and 'Linkage' not in link:
                    linkage[link]=seqid
                    ln_pos.append(position.strip())

        #Generate new output files and remove any existing files
        seg_names=['PB2','PB1','PA','HA','NP','NA','M','NS']
        for seg in seg_names:
            fhandle=seg
            fname="%s_%s.fas" % (cdcid,seg)
            try:
                #os.system('rm ' +fname)
                os.remove(fname)
            except OSError:
                pass
            fhandle=open(fname,'a')

        #consensus FASTA sequence
        fh=open(con_file)
        faiter=(x[1] for x in groupby(fh, lambda line: line[0] == ">"))
        for header in faiter:
            header=header.next()[1:].strip()
            seq="".join(s.strip() for s in faiter.next())
            con_fasta[header]=list(seq)
            variant[header]=header+"_consensus_sequence"+"&"+''.join(seq)
        clone_con_fasta=copy.deepcopy(con_fasta)

        #Get variants that are not linked
        for mk,mb in map_seq.items():
            mv,ref_len=mb.split('=')
            data=mk.split('=')
            pos=''
            clone_con_fasta=copy.deepcopy(con_fasta)
            if data[1] not in ln_pos:
                if not data[4]:
                    try:
                        pos=int(data[1])-1
                        end_pos=pos+int(ref_len)
                        if '-' in mv:
                            rem=[]
                            for i in range(pos,end_pos+1):
                                rem.append('-')
                            clone_con_fasta[data[0]][pos:end_pos]=''.join(rem)
                        clone_con_fasta[data[0]][pos:end_pos]=mv
                    except:
                        pass
                if data[4]:
                    try:
                        pos=int(data[1])-1
                        clone_con_fasta[data[0]][pos]=mv+clone_con_fasta[data[0]][pos]
                    except:
                        pass
            if  pos:
                variant_key=">%s_%s_%s" % (data[0],data[1],data[3])
                variant[variant_key]=variant_key+"&"+''.join(clone_con_fasta[data[0]])


        #Get variants that are linked
        for lk,lv in linkage.items():
            positions=[]
            ln_seq=""
            ln_id=""
            clone_con_fasta=copy.deepcopy(con_fasta)
            for mk,mb in map_seq.items():
                mv,ref_len=mb.split('=')
                data=mk.split('=')
                pos=''
                if lk == data[2] and lv == data[0]:
                    ln_id=data[0]
                    ln_frequency=data[3]
                    if not data[4]:
                        try:
                            pos=int(data[1])-1
                            end_pos=pos+int(ref_len)
                            if '-' in mv:
                                rem=[]
                                for i in range(pos,end_pos+1):
                                    rem.append('-')
                                clone_con_fasta[data[0]][pos:end_pos]=''.join(rem)

                            clone_con_fasta[data[0]][pos:end_pos]=mv
                            positions.append(data[1])
                        except:
                            pass
                    if data[4]:
                        try:
                            pos=int(data[1])-1
                            clone_con_fasta[data[0]][pos]=mv+clone_con_fasta[data[0]][pos]
                            positions.append(data[1])
                        except:
                            pass
                    try:
                        ln_seq=clone_con_fasta[data[0]]
                    except:
                        pass


            if positions and ln_seq:
                variant_key=">%s_%s_%s" % (ln_id,"_".join(sorted(positions)),ln_frequency)
                variant[lk]=variant_key+"&"+''.join(ln_seq)


        #Write sequences to the output files
        for vk,h in variant.items():
            fheader,seq=h.split('&')
            fasta=fheader.split('_')
            #new_header="%s_%s" % (fasta[0],fasta[1])
            new_header="%s" % (fasta[0])
            new_header=re.sub('>','',new_header)
            if 'HA' in fheader:
                out_fname="%s_HA.fas" % (new_header)
                with open(out_fname,'a') as fha:
                    fha.write(format_fasta(re.sub('>','',fheader),''.join(seq),length=50)+"\n")
            if 'PB2' in fheader:
                out_fname="%s_PB2.fas" % (new_header)
                with open(out_fname,'a') as fpb2:
                    fpb2.write(format_fasta(re.sub('>','',fheader),''.join(seq),length=50)+"\n")
            if 'PB1' in fheader:
                out_fname="%s_PB1.fas" % (new_header)
                with open(out_fname,'a') as fpb1:
                    fpb1.write(format_fasta(re.sub('>','',fheader),''.join(seq),length=50)+"\n")
            if 'PA' in fheader:
                out_fname="%s_PA.fas" % (new_header)
                with open(out_fname,'a') as fpa:
                    fpa.write(format_fasta(re.sub('>','',fheader),''.join(seq),length=50)+"\n")
            if 'NP' in fheader:
                out_fname="%s_NP.fas" % (new_header)
                with open(out_fname,'a') as fnp:
                    fnp.write(format_fasta(re.sub('>','',fheader),''.join(seq),length=50)+"\n")
            if 'M' in fheader:
                out_fname="%s_M.fas" % (new_header)
                with open(out_fname,'a') as fm:
                    fm.write(format_fasta(re.sub('>','',fheader),''.join(seq),length=50)+"\n")
            if 'NA' in fheader:
                out_fname="%s_NA.fas" % (new_header)
                with open(out_fname,'a') as fna:
                    fna.write(format_fasta(re.sub('>','',fheader),''.join(seq),length=50)+"\n")
            if 'NS' in fheader:
                out_fname="%s_NS.fas" % (new_header)
                with open(out_fname,'a') as fns:
                    fns.write(format_fasta(re.sub('>','',fheader),''.join(seq),length=50)+"\n")

        fh.close()
        self.dialog.Destroy()
        wx.MessageBox(' The output FASTA files are written to the current working directory. ','Job completion notification',wx.OK | wx.ICON_INFORMATION)



    def OnExit(self, event):
        """Explicitly destroy this frame"""
        self.Destroy()
        self.app.Exit()


class VarMapApp(wx.App):
    """The wx gui application manager for the swap plate app"""
    def OnInit(self):
        """Required setup for the applicaiton"""
        wx.InitAllImageHandlers()

        # set a top level frame and show it
        main_frame = VarMapMainFrame(None, -1, "")
        self.SetTopWindow(main_frame)
        main_frame.Show()

        # must retrun True or the applicaiton will exit
        return True


if __name__ == "__main__":

    def init_gui():
        """Run the swap_plate gui interface"""
        app = VarMapApp(redirect=False)
        app.MainLoop()

    init_gui()
